@@ -5,8 +5,12 @@
 #include "pgrn-global.h"
 #include "pgrn-query-expand.h"
 
+#include <access/genam.h>
 #include <access/heapam.h>
 #include <access/relscan.h>
+#ifdef PGRN_SUPPORT_TABLEAM
+#	include <access/tableam.h>
+#endif
 #include <catalog/pg_operator.h>
 #include <catalog/pg_type.h>
 #include <utils/array.h>
@@ -26,6 +30,9 @@ typedef struct {
 	Form_pg_attribute synonymsAttribute;
 	Snapshot snapshot;
 	IndexScanDesc scan;
+#ifdef PGRN_SUPPORT_TABLEAM
+	TupleTableSlot *slot;
+#endif
 	StrategyNumber scanStrategy;
 	Oid scanOperator;
 	RegProcedure scanProcedure;
@@ -50,7 +57,7 @@ func_query_expander_postgresql(grn_ctx *ctx,
 	Datum scanKeyDatum;
 	ScanKeyData scanKeys[1];
 	int nKeys = 1;
-	HeapScanDesc heapScan = NULL;
+	PGrnTableScanDesc heapScan = NULL;
 	HeapTuple tuple;
 	int ith_synonyms = 0;
 
@@ -107,10 +114,10 @@ func_query_expander_postgresql(grn_ctx *ctx,
 					InvalidStrategy,
 					currentData.scanProcedure,
 					scanKeyDatum);
-		heapScan = heap_beginscan(currentData.table,
-								  currentData.snapshot,
-								  nKeys,
-								  scanKeys);
+		heapScan = pgrn_table_beginscan(currentData.table,
+										currentData.snapshot,
+										nKeys,
+										scanKeys);
 	}
 
 	while (true)
@@ -119,9 +126,22 @@ func_query_expander_postgresql(grn_ctx *ctx,
 		bool isNULL;
 
 		if (currentData.scan)
+		{
+#ifdef PGRN_SUPPORT_TABLEAM
+			if (index_getnext_slot(currentData.scan,
+								   ForwardScanDirection,
+								   currentData.slot))
+				tuple = ExecFetchSlotHeapTuple(currentData.slot, false, NULL);
+			else
+				tuple = NULL;
+#else
 			tuple = index_getnext(currentData.scan, ForwardScanDirection);
+#endif
+		}
 		else
+		{
 			tuple = heap_getnext(heapScan, ForwardScanDirection);
+		}
 
 		if (!tuple)
 			break;
@@ -466,10 +486,16 @@ pgroonga_query_expand(PG_FUNCTION_ARGS)
 										   currentData.snapshot,
 										   nKeys,
 										   nOrderBys);
+#ifdef PGRN_SUPPORT_TABLEAM
+		currentData.slot = table_slot_create(currentData.table, NULL);
+#endif
 	}
 	else
 	{
 		currentData.scan = NULL;
+#ifdef PGRN_SUPPORT_TABLEAM
+		currentData.slot = NULL;
+#endif
 	}
 
 	currentData.scanProcedure = get_opcode(currentData.scanOperator);
@@ -488,6 +514,9 @@ pgroonga_query_expand(PG_FUNCTION_ARGS)
 	}
 	if (currentData.scan)
 	{
+#ifdef PGRN_SUPPORT_TABLEAM
+		ExecDropSingleTupleTableSlot(currentData.slot);
+#endif
 		index_endscan(currentData.scan);
 		index_close(index, NoLock);
 	}
